# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import copy
import time
import subprocess

import odatix.lib.printc as printc
from odatix.lib.utils import *

script_name = os.path.basename(__file__)


def check_tool(tool, script_path, makefile, rule, supported_tools):
  print('checking the selected eda tool "' + tool + '" ..', end="")
  sys.stdout.flush()
  tool_makefile_file = script_path + "/" + tool + "/" + makefile
  test_process = subprocess.Popen(
    ["make", "-f", tool_makefile_file, rule, "--no-print-directory"], stdout=subprocess.DEVNULL, stderr=subprocess.PIPE
  )
  while test_process.poll() is None:
    print(".", end="", flush=True)
    time.sleep(0.5)
  if test_process.returncode == 0:
    printc.green(" success!")
  else:
    printc.red(" failed!")
    error_message = test_process.stderr.read().decode()
    printc.error('Could not launch eda tool "' + tool + '"', script_name)
    printc.cyan("error details: ", script_name, end="")
    printc.red(error_message, end="")
    printc.note("Did you add the tool path to your PATH environment variable?", script_name)
    printc.note("Example -> PATH=$PATH:/opt/xilinx/2022/Vivado/2022.2/bin", script_name)
    if tool not in supported_tools:
      printc.note(
        'The selected eda tool "{}" is not one of the supported tool. '.format(tool)
        + "Check out Odatix's documentation to add support for your own eda tool",
        script_name,
      )
      printc.note('Make sure there is a valid rule "' + rule + '" in "' + tool_makefile_file + '"', script_name)
    sys.exit(-1)
  print()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import yaml

import odatix.lib.printc as printc
from odatix.lib.utils import *

script_name = os.path.basename(__file__)

def get_synth_settings(settings_filename):
  # failsafe
  if settings_filename is None:
    printc.error("No settings file specified: get_synth_settings(settings_filename) -> settings_filename is None", script_name)
    sys.exit(-1)

  # get synth settings from yaml file
  if not os.path.isfile(settings_filename):
    printc.error("Settings file \"" + settings_filename + "\" does not exist", script_name)
    sys.exit(-1)

  with open(settings_filename, 'r') as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)
    try:
      overwrite       = read_from_list("overwrite", settings_data, settings_filename, type=bool, script_name=script_name)
      ask_continue    = read_from_list("ask_continue", settings_data, settings_filename, type=bool, script_name=script_name)
      show_log_if_one = read_from_list("show_log_if_one", settings_data, settings_filename, type=bool, script_name=script_name)
      nb_jobs         = read_from_list("nb_jobs", settings_data, settings_filename, type=int, script_name=script_name)
      architectures   = read_from_list("architectures", settings_data, settings_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      sys.exit(-1) # if a key is missing
  return overwrite, ask_continue, show_log_if_one, nb_jobs, architectures


def get_sim_settings(settings_filename):
  # failsafe
  if settings_filename is None:
    printc.error("No settings file specified: get_sim_settings(settings_filename) -> settings_filename is None", script_name)
    sys.exit(-1)

  # get sim settings from yaml file
  if not os.path.isfile(settings_filename):
    printc.error("Settings file \"" + settings_filename + "\" does not exist", script_name)
    sys.exit(-1)

  with open(settings_filename, 'r') as f:
    settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    try:
      overwrite       = read_from_list("overwrite", settings_data, settings_filename, type=bool, script_name=script_name)
      ask_continue    = read_from_list("ask_continue", settings_data, settings_filename, type=bool, script_name=script_name)
      show_log_if_one = read_from_list("show_log_if_one", settings_data, settings_filename, type=bool, script_name=script_name)
      nb_jobs         = read_from_list("nb_jobs", settings_data, settings_filename, type=int, script_name=script_name)
      simulations     = read_from_list("simulations", settings_data, settings_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      sys.exit(-1) # if a key is missing
  return overwrite, ask_continue, show_log_if_one, nb_jobs, simulations



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

class colors:
  GREY = "\033[90m"
  RED = "\033[91m"
  GREEN = "\033[92m"
  YELLOW = "\033[93m"
  BLUE = "\033[94m"
  MAGENTA = "\033[95m"
  CYAN = "\033[96m"
  ENDC = "\033[0m"
  BOLD = "\033[1m"
  BLINK = "\033[5m"
  UNDERLINE = "\033[4m"

def say(message="", script_name="", color=colors.ENDC, end="\n"):
  if script_name != "":
    script_name = colors.GREY + "[" + script_name + "]" + colors.ENDC + " "
  print(script_name + color + message + colors.ENDC, end=end)

def error(message, script_name="", end="\n"):
  if script_name != "":
    script_name = colors.GREY + "[" + script_name + "]" + colors.ENDC + " "
  print(script_name + colors.BOLD + colors.RED + "error" + colors.ENDC + colors.RED + ": " + message + colors.ENDC, end=end)

def internal_error(message, script_name="", end="\n"):
  if script_name != "":
    script_name = colors.GREY + "[" + script_name + "]" + colors.ENDC + " "
  print(script_name + colors.BOLD + colors.RED + "internal error" + colors.ENDC + colors.RED + ": " + message + colors.ENDC, end=end)

def warning(message, script_name="", end="\n"):
  if script_name != "":
    script_name = colors.GREY + "[" + script_name + "]" + colors.ENDC + " "
  print(script_name + colors.YELLOW + "warning: " + message + colors.ENDC, end=end)

def note(message, script_name="", end="\n"):
  if script_name != "":
    script_name = colors.GREY + "[" + script_name + "]" + colors.ENDC + " "
  print(script_name + colors.CYAN + "note: " + message + colors.ENDC, end=end)

def header(message):
  print(colors.BOLD + colors.CYAN + message + colors.ENDC)

def subheader(message):
  print(colors.CYAN + message + colors.ENDC)

def bold(message, color=colors.ENDC, end="\n"):
  print(color + colors.BOLD + message + colors.ENDC, end=end)

def grey(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.GREY, end=end)

def red(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.RED, end=end)

def green(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.GREEN, end=end)

def yellow(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.YELLOW, end=end)

def blue(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.BLUE, end=end)

def magenta(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.MAGENTA, end=end)

def cyan(message, script_name="", end="\n"):
  say(message=message, script_name=script_name, color=colors.CYAN, end=end)

def color(color=colors.ENDC):
  print(color, end="")

def endc(message="", script_name="", end=""):
  say(message=message, script_name=script_name, color=colors.ENDC, end=end)


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import pandas as pd
import plotly.graph_objs as go

import odatix.explorer.legend as legend
import odatix.explorer.navigation as navigation

page_name = "radar"


def layout(explorer):
  legend_items = legend.create_legend_items(explorer, page_name)

  return html.Div(
    [
      navigation.top_bar(page_name),
      navigation.side_bar(
        content=html.Div(
          id=f"sidebar-content-{page_name}",
          className="sidebar-content-holder",
          children=[
            html.H2("Data"),
            html.Div(
              [
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("YAML File")]),
                    dcc.Dropdown(
                      id="yaml-dropdown",
                      options=[{"label": yaml_file, "value": yaml_file} for yaml_file in explorer.valid_yaml_files],
                      value=explorer.valid_yaml_files[0] if explorer.valid_yaml_files else None,
                    ),
                  ],
                ),
                html.Div(
                  className="title-dropdown",
                  children=[
                    html.Div(className="dropdown-label", children=[html.Label("Target")]),
                    dcc.Dropdown(
                      id=f"target-dropdown-{page_name}",
                      value=explorer.dfs[explorer.valid_yaml_files[0]]["Target"].iloc[0]
                      if explorer.valid_yaml_files
                      else None,
                    ),
                  ],
                ),
              ]
            ),
            html.H2("Architectures"),
            html.Div(
              [
                html.Div(
                  [
                    html.Button("Show All", id="show-all", n_clicks=0),
                    html.Button("Hide All", id="hide-all", n_clicks=0),
                  ]
                ),
                html.Div(legend_items, id="custom-legend", style={"margin-top": "15px", "margin-bottom": "15px"}),
              ],
              style={"display": "inline-block", "margin-left": "20px"},
            ),
            html.H2("Display Settings"),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Legend Mode")]),
                dcc.Dropdown(
                  id="legend-dropdown",
                  options=[
                    {"label": "Hide", "value": "hide_legend"},
                    {"label": "Show", "value": "show_legend"},
                    {"label": "Separate", "value": "separate_legend"},
                  ],
                  value="hide_legend",
                ),
              ],
              style={"margin-bottom": "5px"},
            ),
            html.Div(
              className="toggle-container",
              children=[
                dcc.Checklist(
                  id="toggle-title",
                  options=[{"label": " Show Title", "value": "show_title"}],
                  value=["show_title"],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
                dcc.Checklist(
                  id="toggle-lines",
                  options=[{"label": " Show Lines", "value": "show_lines"}],
                  value=["show_lines"],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
                dcc.Checklist(
                  id="toggle-close-line",
                  options=[{"label": " Close Lines", "value": "close_line"}],
                  value=[""],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
              ],
            ),
            html.H2("Export Settings"),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Download Format")]),
                dcc.Dropdown(
                  id="dl-format-dropdown",
                  options=[
                    {"label": "SVG", "value": "svg"},
                    {"label": "PNG", "value": "png"},
                    {"label": "JPEG", "value": "jpeg"},
                    {"label": "WEBP", "value": "webp"},
                  ],
                  value="svg",
                ),
              ],
              style={"margin-bottom": "5px"},
            ),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Background Color")]),
                dcc.Dropdown(
                  id="background-dropdown",
                  options=[
                    {"label": "Transparent", "value": "rgba(255, 255, 255, 0)"},
                    {"label": "White", "value": "rgba(255, 255, 255, 255)"},
                  ],
                  value="rgba(255, 255, 255, 0)",
                ),
              ],
            ),
          ],
        ),
        page_name=page_name,
      ),
      html.Div(
        id=f"content-{page_name}",
        children=[
          html.Div(id="radar-graphs"),
        ],
        className="content",
        style={
          "marginLeft": navigation.side_bar_width,
          "width": "calc(100%-" + navigation.side_bar_width + ")",
          "height": "100%",
          "justify-content": "center",
        },
      ),
    ]
  )


def make_legend_chart(df, all_architectures, visible_architectures, toggle_title, background, mode):
  fig = go.Figure()

  if not visible_architectures:
    return None

  for i, architecture in enumerate(all_architectures):
    if architecture in visible_architectures:
      df_architecture = df[df["Architecture"] == architecture]
      fig.add_trace(
        go.Scatterpolar(
          r=[None],
          theta=df_architecture["Configuration"],
          mode=mode,
          name=architecture,
          marker_color=legend.get_color(i),
        )
      )

  fig.update_layout(
    polar_bgcolor="rgba(255, 255, 255, 0)",
    paper_bgcolor=background,
    showlegend=True,
    margin=dict(l=60, r=60, t=60, b=60),
    title="Legend" if "show_title" in toggle_title else "",
    title_x=0.5,
    polar=dict(radialaxis=dict(visible=False), angularaxis=dict(visible=False)),
    autosize=True,
    legend_x=0,
    legend_y=1,
    width=475,
    height=475,
  )

  return fig


def make_radar_chart(
  df,
  units,
  metric,
  all_configurations,
  all_architectures,
  visible_architectures,
  legend_dropdown,
  toggle_title,
  toggle_close,
  background,
  mode,
):
  df.loc[:, metric] = pd.to_numeric(df[metric], errors="coerce")
  df = df.dropna(subset=[metric])

  if df.empty:
    fig = go.Figure(data=go.Scatterpolar())
    fig.update_layout(
      polar=dict(radialaxis=dict(visible=False), angularaxis=dict(showticklabels=False)),
      showlegend=False,
      title=metric.replace("_", " ") if metric is not None else "",
      title_x=0.5,
    )
    return fig

  metric_display = metric.replace("_", " ")
  unit = legend.unit_to_html(units.get(metric, ""))

  fig = go.Figure(
    data=go.Scatterpolar(
      r=[None for c in all_configurations],
      theta=all_configurations,
      marker_color="rgba(0, 0, 0, 0)",
      showlegend=False,
    )
  )

  for i, architecture in enumerate(all_architectures):
    if architecture in visible_architectures:
      df_architecture = df[df["Architecture"] == architecture]
      if "close_line" in toggle_close:
        first_row = df_architecture.iloc[0:1]
        df_architecture = df_architecture._append(first_row, ignore_index=True)

      fig.add_trace(
        go.Scatterpolar(
          r=df_architecture[metric],
          theta=df_architecture["Configuration"],
          mode=mode,
          name=architecture,
          marker_color=legend.get_color(i),
          hovertemplate="<br>".join(
            [
              "Architecture: %{fullData.name}",
              "Configuration: %{theta}",
              str(metric_display) + ": %{r} " + unit,
              "<extra></extra>",
            ]
          ),
        )
      )

  fig.update_layout(
    # template='plotly_dark',
    paper_bgcolor=background,
    polar=dict(radialaxis=dict(visible=True, range=[0, df[metric].max() if not df[metric].empty else 1])),
    showlegend="show_legend" in legend_dropdown,
    margin=dict(l=60, r=60, t=60, b=60),
    title=metric_display if "show_title" in toggle_title else None,
    title_x=0.5,
    width=840 if "show_legend" in legend_dropdown else 475,
    height=475,
  )

  return fig


def make_figure_div(fig, filename, dl_format, remove_zoom=False):
  if fig is not None:
    to_remove = ["lasso", "select"]
    if remove_zoom:
      to_remove.append("zoom")
    return html.Div(
      [
        dcc.Graph(
          figure=fig,
          style={"width": "100%"},
          config={
            "displayModeBar": True,
            "displaylogo": False,
            "modeBarButtonsToRemove": to_remove,
            "toImageButtonOptions": {"format": dl_format, "scale": "3", "filename": filename},
          },
        )
      ],
      style={"flex": "0 0 auto", "margin": "0px"},
    )
  else:
    return html.Div([html.Div([], style={"width": "475px"})], style={"flex": "0 0 auto", "margin": "0px"})


def make_all_radar_charts(
  df,
  units,
  metrics,
  all_configurations,
  all_architectures,
  visible_architectures,
  yaml_name,
  legend_dropdown,
  toggle_title,
  toggle_lines,
  toggle_close,
  dl_format,
  background,
):
  radar_charts = []

  mode = "lines+markers" if "show_lines" in toggle_lines else "markers"

  for metric in metrics:
    fig = make_radar_chart(
      df,
      units,
      metric,
      all_configurations,
      all_architectures,
      visible_architectures,
      legend_dropdown,
      toggle_title,
      toggle_close,
      background,
      mode,
    )
    filename = "Odatix-{}-{}-{}".format(yaml_name, page_name, metric)
    radar_charts.append(make_figure_div(fig, filename, dl_format))

  # Add legend chart
  if "separate_legend" in legend_dropdown:
    legend_fig = make_legend_chart(df, all_architectures, visible_architectures, toggle_title, background, mode)
    radar_charts.append(
      make_figure_div(legend_fig, "Odatix-" + str(page_name) + "-legend", dl_format, remove_zoom=True)
    )

  return radar_charts


def setup_callbacks(explorer):
  @explorer.app.callback(
    Output("radar-graphs", "children"),
    [
      Input("yaml-dropdown", "value"),
      Input(f"target-dropdown-{page_name}", "value"),
      Input("show-all", "n_clicks"),
      Input("hide-all", "n_clicks"),
      Input("legend-dropdown", "value"),
      Input("toggle-title", "value"),
      Input("toggle-lines", "value"),
      Input("toggle-close-line", "value"),
      Input("dl-format-dropdown", "value"),
      Input("background-dropdown", "value"),
    ]
    + [Input(f"checklist-{architecture}-{page_name}", "value") for architecture in explorer.all_architectures],
  )
  def update_radar_charts(
    selected_yaml,
    selected_target,
    show_all,
    hide_all,
    legend_dropdown,
    toggle_title,
    toggle_lines,
    toggle_close,
    dl_format,
    background,
    *checklist_values,
  ):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return html.Div(className="error", children=[html.Div("Please select a YAML file.")])

    if not selected_target:
      return html.Div(className="error", children=[html.Div("Please select a target.")])

    ctx = dash.callback_context
    triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

    if triggered_id in ["show-all", "hide-all"]:
      visible_architectures = set(explorer.all_architectures if triggered_id == "show-all" else [])
    else:
      visible_architectures = set(
        architecture for i, architecture in enumerate(explorer.all_architectures) if checklist_values[i]
      )

    filtered_df = explorer.dfs[selected_yaml][
      (explorer.dfs[selected_yaml]["Target"] == selected_target)
      & (explorer.dfs[selected_yaml]["Architecture"].isin(visible_architectures))
    ]

    unique_configurations = sorted(filtered_df["Configuration"].unique())

    metrics = [col for col in filtered_df.columns if col not in ["Target", "Architecture", "Configuration"]]
    all_configurations = explorer.all_configurations
    all_architectures = explorer.all_architectures

    for metric in metrics:
      filtered_df.loc[:, metric] = pd.to_numeric(filtered_df[metric], errors="coerce")

    yaml_name = os.path.splitext(selected_yaml)[0] + "-" + selected_target

    radar_charts = make_all_radar_charts(
      filtered_df,
      explorer.units[selected_yaml],
      metrics,
      unique_configurations,
      all_architectures,
      visible_architectures,
      yaml_name,
      legend_dropdown,
      toggle_title,
      toggle_lines,
      toggle_close,
      dl_format,
      background,
    )

    return html.Div(radar_charts, style={"display": "flex", "flex-wrap": "wrap", "justify-content": "space-evenly"})

  @explorer.app.callback(Output(f"target-dropdown-{page_name}", "options"), Input("yaml-dropdown", "value"))
  def update_dropdowns_radar(selected_yaml):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return [], []

    df = explorer.dfs[selected_yaml]
    available_targets = [{"label": target, "value": target} for target in df["Target"].unique()]

    return available_targets

  legend.setup_callbacks(explorer, page_name)
  navigation.setup_sidebar_callbacks(explorer, page_name)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import yaml
import argparse

from odatix.lib.settings import OdatixSettings
import odatix.lib.re_helper as rh
import odatix.lib.printc as printc

######################################
# Settings
######################################

DEFAULT_BENCHMARK = "dhrystone"
DEFAULT_SIM_FILE = "log/sim.log"

status_done = 'Done: 100%'

script_name = os.path.basename(__file__)

dmips_per_mhz_pattern = re.compile("(.*)DMIPS_Per_MHz: ([0-9.]+)")
bad_value = None

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument('-b', '--benchmark', default=DEFAULT_BENCHMARK, help='benchmark to parse (default: ' + DEFAULT_BENCHMARK + ')')
  parser.add_argument('-S', '--sim_file', default=DEFAULT_SIM_FILE, help='simulation log file (default: ' + DEFAULT_SIM_FILE + ')')
  parser.add_argument('-B', '--benchmark_file', help='output benchmark file')
  parser.add_argument('-w', '--work', help='simulation work directory')
  parser.add_argument('-c', '--config', default=OdatixSettings.DEFAULT_SETTINGS_FILE, help='global settings file for Odatix (default: ' + OdatixSettings.DEFAULT_SETTINGS_FILE + ')')

def parse_arguments():
  parser = argparse.ArgumentParser(description='Process benchmark results')
  add_arguments(parser)
  return parser.parse_args()

######################################
# Misc functions
######################################

def corrupted_directory(variant):
  printc.warning(variant + " => synthesis has not finished or directory has been corrupted", script_name)

def safe_cast(val, to_type, default=None):
  try:
      return to_type(val)
  except (ValueError, TypeError):
      return default

######################################
# Parsing functions
######################################

def get_dmips_per_mhz(file):
  return rh.get_re_group_from_file(file, dmips_per_mhz_pattern, group_id=2)

######################################
# Format functions
######################################

def cast_to_int(input):
  if input == bad_value:
    return '/'
  else:
    return safe_cast(input, int, 0)

def cast_to_float(input):
  if input == bad_value:
    return '/'
  else:
    return safe_cast(input, float, 0.0)

def write_to_yaml(input, sim_file, output_file):
  yaml_data = {}

  for arch in sorted(next(os.walk(input))[1]):
    yaml_data[arch] = {}
    for variant in sorted(next(os.walk(os.path.join(input, arch)))[1]):
      cur_path = os.path.join(input, arch, variant)
      cur_sim_file = os.path.join(cur_path, sim_file)

      if not os.path.exists(cur_sim_file):
        corrupted_directory(arch + '/' + variant)
        continue

      dmips_per_mhz = get_dmips_per_mhz(cur_sim_file)
      dmips_per_mhz = cast_to_float(dmips_per_mhz) if dmips_per_mhz != ' /  ' else ""
      yaml_data[arch][variant] = {
        'DMIPS_per_MHz': dmips_per_mhz,
      }

  output_path = os.path.dirname(output_file)
  if not os.path.exists(output_path):
    os.makedirs(output_path, exist_ok=True)
  try:
    with open(output_file, 'w') as file:
      yaml.dump(yaml_data, file, default_flow_style=False, sort_keys=False)
      printc.say("Results written to \"" + output_file + "\"", script_name=script_name)
  except Exception as e:
    printc.error("Could not write results to \"" + output_file + "\"", script_name=script_name)
    printc.cyan("error details: ", end="", script_name=script_name)
    print(str(e))

######################################
# Export Results
######################################

def export_benchmark(input, output, benchmark, sim_file):
  print(printc.colors.CYAN + "Export " +  benchmark + " results" + printc.colors.ENDC)
  input = os.path.join(input, benchmark)
  if not os.path.isdir(input):
    printc.error("input directory \"" + input + "\" does not exist", script_name)
    sys.exit(1)

  #if format in ['yml']:
  write_to_yaml(input, sim_file, output)

  print()

######################################
# Main
######################################

def main(args, settings=None):

  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.benchmark_file is not None:
    benchmark_file  = args.benchmark_file
  else:
    benchmark_file = settings.benchmark_file

  if args.work is not None:
    work = args.work
  else:
    work = settings.sim_work_path

  export_benchmark(
    input=work,
    output=benchmark_file,
    benchmark=args.benchmark,
    sim_file=args.sim_file
  )

if __name__ == "__main__":
  args = parse_arguments()
  main(args)


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import pandas as pd
import yaml

import odatix.explorer.page_xy as page_xy
import odatix.explorer.page_vs as page_vs
import odatix.explorer.page_radar as page_radar

import odatix.lib.printc as printc

script_name = os.path.basename(__file__)


class ResultExplorer:
  def __init__(self, result_path="results", yaml_prefix="results_"):
    self.result_path = result_path
    self.yaml_prefix = yaml_prefix

    # Check paths
    if not os.path.exists(result_path):
      printc.error('Could not find result path "' + result_path + '"')
      sys.exit(-1)

    # Initialize additional instance variables here
    self.yaml_files = [
      file for file in os.listdir(self.result_path) if file.endswith(".yml") and file.startswith(yaml_prefix)
    ]
    self.valid_yaml_files = []
    self.all_data = {}
    self.dfs = {}
    self.units = {}

    # Load and validate YAML files
    self.load_yaml_files()

    if not self.valid_yaml_files:
      printc.error(
        'Could not find any valid YAML file in "' + self.result_path + '", exiting.', script_name=script_name
      )
      sys.exit(-1)

    self.all_architectures = sorted(
      set(architecture for df in self.dfs.values() for architecture in df["Architecture"].unique())
    )
    self.all_configurations = sorted(set(config for df in self.dfs.values() for config in df["Configuration"].unique()))

    self.app = dash.Dash(__name__)
    self.app.title = "Odatix"
    self.setup_layout()
    self.setup_callbacks()

  def load_yaml_files(self):
    """
    Load and validate YAML files from the specified path.
    """
    for yaml_file in self.yaml_files:
      file_path = os.path.join(self.result_path, yaml_file)
      try:
        data, units = self.get_yaml_data(file_path)
        df = self.update_dataframe(data)
        if df is None:
          printc.warning('YAML file  "' + yaml_file + '" is empty or corrupted, skipping...', script_name=script_name)
          printc.note(
            'Run fmax synthesis with the correct settings to generate "' + yaml_file + '"', script_name=script_name
          )
        else:
          self.all_data[yaml_file] = data
          self.units[yaml_file] = units
          self.valid_yaml_files.append(yaml_file)
          self.dfs[yaml_file] = df
      except:
        printc.warning(
          'YAML file  "' + yaml_file + '" is not a valid result file, skipping...', script_name=script_name
        )

  def get_yaml_data(self, file_path):
    """
    Load YAML data from a file.
    """
    with open(file_path, "r") as file:
      yaml_content = yaml.safe_load(file)
      units = yaml_content.get("units", {})
      synth_results = yaml_content.get("fmax_results", {})
      return synth_results, units

  def update_dataframe(self, yaml_data):
    """
    Update the dataframe with YAML data.
    """
    data = []
    for target, architectures in yaml_data.items():
      for architecture, configurations in architectures.items():
        for config, metrics in configurations.items():
          row = metrics.copy()
          row["Target"] = target
          row["Architecture"] = architecture
          row["Configuration"] = config
          data.append(row)
    df = pd.DataFrame(data)

    # Check if the dataframe contains the required columns and they are not empty
    required_columns = ["Target", "Architecture", "Configuration"]
    if not all(column in df.columns and not df[column].empty for column in required_columns):
      return None
    return df

  def update_metrics(self, yaml_data):
    """
    Update metrics based on YAML data.
    """
    metrics_from_yaml = set()
    for target_data in yaml_data.values():
      for architecture_data in target_data.values():
        for configuration_data in architecture_data.values():
          metrics_from_yaml.update(configuration_data.keys())
    return metrics_from_yaml

  def setup_layout(self):
    """
    Setup the layout of the Dash application.
    """
    self.app.layout = html.Div(
      [dcc.Location(id="url", refresh=False), html.Div([html.Div(id="page-content", className="container")])]
    )

  def setup_callbacks(self):
    """
    Setup Dash callbacks for interactivity.
    """

    @self.app.callback(Output("page-content", "children"), [Input("url", "pathname")])
    def display_page(pathname):
      if pathname == "/xy":
        return page_xy.layout(self)
      elif pathname == "/vs":
        return page_vs.layout(self)
      elif pathname == "/radar":
        return page_radar.layout(self)
      else:
        return page_xy.layout(self)

    page_xy.setup_callbacks(self)
    page_vs.setup_callbacks(self)
    page_radar.setup_callbacks(self)

  def run(self):
    self.app.run_server(debug=False)


if __name__ == "__main__":
  explorer = ResultExplorer()
  explorer.run()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import argparse
import subprocess

import odatix.lib.printc as printc
from odatix.lib.replace_params import replace_params
from odatix.lib.parallel_job_handler import ParallelJobHandler, ParallelJob
from odatix.lib.settings import OdatixSettings
from odatix.lib.simulation_handler import SimulationHandler
from odatix.lib.utils import read_from_list, copytree, create_dir, ask_to_continue
from odatix.lib.prepare_work import edit_config_file
from odatix.lib.check_tool import check_tool
from odatix.lib.run_settings import get_sim_settings

######################################
# Settings
######################################

work_path = "work/simulation"
work_script_path = "scripts"
common_script_path = "_common"
log_path = "log"
arch_path = "architectures"
sim_path = "simulations"
rtl_path = "rtl"

nb_jobs = 4

param_settings_filename = "_settings.yml"
sim_settings_filename = "_settings.yml"
sim_makefile_filename = "Makefile"
sim_rule = "sim"
sim_progress_filename = "progress.log"

progress_bar_size = 50
refresh_time = 5

default_fmax_lower_bound = 50
default_fmax_upper_bound = 500

sim_status_pattern = re.compile(r"(.*): ([0-9]+)%(.*)")

script_name = os.path.basename(__file__)

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument('-o', '--overwrite', action='store_true', help='overwrite existing results')
  parser.add_argument('-y', '--noask', action='store_true', help='do not ask to continue')
  parser.add_argument('-i', '--input', help='input settings file')
  parser.add_argument('-a', '--archpath', help='architecture directory')
  parser.add_argument('-s', '--simpath', help='simulation directory')
  parser.add_argument('-w', '--work', help='simulation work directory')
  parser.add_argument('-c', '--config', default=OdatixSettings.DEFAULT_SETTINGS_FILE, help='global settings file for Odatix (default: ' + OdatixSettings.DEFAULT_SETTINGS_FILE + ')')

def parse_arguments():
  parser = argparse.ArgumentParser(description='Run parallel simulations')
  add_arguments(parser)
  return parser.parse_args()


######################################
# Run Simulations
######################################

def run_simulations(run_config_settings_filename, arch_path, sim_path, work_path, overwrite, noask):
  _overwrite, ask_continue, show_log_if_one, nb_jobs, simulations = get_sim_settings(run_config_settings_filename)

  if simulations is None:
    printc.error('The "simulations" section of "' + run_config_settings_filename + '" is empty.', script_name)
    printc.note('You must define your simulations in "' + run_config_settings_filename + '" before using this command.', script_name)
    printc.note("Check out examples Odatix's documentation for more information.", script_name)
    sys.exit(-1)

  if overwrite:
    overwrite = True
  else:
    overwrite = _overwrite
  
  if noask:
    ask_continue = False

  ParallelJob.set_patterns(sim_status_pattern)

  sim_handler = SimulationHandler(
    work_path = work_path,
    arch_path = arch_path,
    sim_path = sim_path,
    work_script_path = work_script_path,
    log_path = log_path,
    overwrite = overwrite,
    param_settings_filename = param_settings_filename,
    sim_settings_filename = sim_settings_filename,
    sim_makefile_filename = sim_makefile_filename
  )

  if simulations is None:
    printc.note("No simulation selected. Exiting.", script_name)
    sys.exit(-1)
    
  try:
    simulation_instances = sim_handler.get_simulations(simulations)
  except Exception as e:
    printc.error("Could not get list \"simulations\" from \"" + run_config_settings_filename + "\".", script_name=script_name)
    printc.note("Is the YAML file valid? Are you missing a ':'? Is the indentation correct?", script_name=script_name)
    printc.cyan("error details: ", end="", script_name=script_name)
    print(str(e))
    sys.exit(-1)

  # print checklist summary
  sim_handler.print_summary()

  # ask to quit or continue
  if sim_handler.get_valid_sim_count() > 0:
    if ask_continue:
      print()
      ask_to_continue()
  else:
    sys.exit(-1)

  print()

  job_list = []

  def prepare_job(sim_instance):
    
    if True:
      # create directory
      create_dir(sim_instance.tmp_dir)

      # copy simulation sources
      copytree(sim_instance.source_sim_dir, sim_instance.tmp_dir, dirs_exist_ok = True)
     
      # copy design 
      if sim_instance.architecture.design_path != -1:
        try:
          copytree(sim_instance.architecture.design_path, sim_instance.tmp_dir, dirs_exist_ok = True)
        except:
          printc.error("Could not copy \"" + sim_instance.architecture.design_path + "\" into work directory \"" + sim_instance.tmp_dir + "\"", script_name)
          printc.note("make sure there are no file or folder named identically in the two directories", script_name)
          return

      # copy rtl (if exists) 
      if not sim_instance.architecture.generate_rtl:
        copytree(sim_instance.architecture.rtl_path, sim_instance.tmp_dir + '/' + 'rtl', dirs_exist_ok = True)

      # replace parameters
      if sim_instance.architecture.use_parameters:
        #printc.subheader("Replace parameters")
        param_target_file = sim_instance.tmp_dir + '/' + sim_instance.architecture.param_target_filename
        param_filename = arch_path + '/' + sim_instance.architecture.arch_name + '.txt'
        replace_params(
          base_text_file=param_target_file, 
          replacement_text_file=param_filename, 
          output_file=param_target_file, 
          start_delimiter=sim_instance.architecture.start_delimiter, 
          stop_delimiter=sim_instance.architecture.stop_delimiter, 
          replace_all_occurrences=False,
          silent=True
        )
        #print()

      # replace parameters again (override)
      if sim_instance.override_parameters:
        #printc.subheader("Replace parameters")
        param_target_file = sim_instance.tmp_dir + '/' + sim_instance.override_param_target_filename
        param_file = sim_instance.tmp_dir + '/' + sim_instance.override_param_filename
        replace_params(
          base_text_file=param_target_file, 
          replacement_text_file=param_file, 
          output_file=param_target_file, 
          start_delimiter=sim_instance.override_start_delimiter, 
          stop_delimiter=sim_instance.override_stop_delimiter, 
          replace_all_occurrences=False,
          silent=True
        )

      # run simulation command
      command = (
        "make {}".format(sim_rule)
        + ' RTL_DIR="{}"'.format(rtl_path)
        + ' ODATIX_DIR="{}"'.format(OdatixSettings.odatix_path)
        + ' LOG_DIR="{}"'.format(os.path.realpath(os.path.join(sim_instance.tmp_dir, log_path)))
        + ' CLOCK_SIGNAL="{}"'.format(sim_instance.architecture.clock_signal)
        + ' TOP_LEVEL_MODULE="{}"'.format(sim_instance.architecture.top_level_module)
        + " --no-print-directory"
      )

      sim_progress_file = os.path.join(sim_instance.tmp_dir, log_path, sim_progress_filename)

      running_sim = ParallelJob(
        process=None,
        command=command,
        directory=sim_instance.tmp_dir,
        generate_rtl=sim_instance.architecture.generate_rtl,
        generate_command=sim_instance.architecture.generate_command,
        target="",
        arch="",
        display_name=sim_instance.sim_display_name,
        status_file="",
        progress_file=sim_progress_file,
        tmp_dir=sim_instance.tmp_dir,
        status="idle",
      )

      job_list.append(running_sim)

  for sim_instance in simulation_instances:
    prepare_job(sim_instance)

  parallel_jobs = ParallelJobHandler(
    job_list=job_list,
    nb_jobs=nb_jobs,
    process_group=True,
  )
  job_exit_success = parallel_jobs.run()

######################################
# Main
######################################

def main(args, settings=None):

  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.input is not None:
    run_config_settings_filename  = args.input
  else:
    run_config_settings_filename = settings.simulation_settings_file

  if args.archpath is not None:
    arch_path = args.archpath
  else:
    arch_path = settings.arch_path

  if args.simpath is not None:
    sim_path = args.simpath
  else:
    sim_path = settings.sim_path

  if args.work is not None:
    work_path = args.work
  else:
    work_path = settings.sim_work_path

  overwrite = args.overwrite
  noask = args.noask

  run_simulations(run_config_settings_filename, arch_path, sim_path, work_path, overwrite, noask)

if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import json
import time
import argparse
import subprocess

# add local libs to path
current_dir = os.path.dirname(os.path.abspath(__file__))
lib_path = os.path.join(current_dir, '../../../scripts/lib')
sys.path.append(lib_path)

import printc

from architecture_handler import Architecture

######################################
# Settings
######################################

config_filename = "config.json"
yaml_config_filename = "settings.yml"
generate_command = "/openlane/flow.tcl -init_design_config"

script_name = "eda_tools/openlane/scripts/" + os.path.basename(__file__)

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument('-b', '--basepath', default='.', help='base path (default: .)')
  parser.add_argument('-c', '--clock', help='clock signal')
  parser.add_argument('-d', '--docker', help='docker id or name')
  parser.add_argument('-n', '--name', help='design name')

def parse_arguments():
  parser = argparse.ArgumentParser(description='Edit OpenLane config.json file')
  add_arguments(parser)
  return parser.parse_args()

######################################
# Main
######################################

def main(base_path, clock_signal, docker_id, design_name):
  config_file = os.path.join(base_path, config_filename)

  # Check if config file exists
  if not os.path.isfile(config_file):
    try:
      #command = "docker exec -it " + docker_id + " /bin/sh -c ' cd " + base_path + "; " + generate_command + " --design-dir " + base_path + "'"
      command = "docker exec -it " + docker_id + " /bin/sh -c ' python3 /openlane/scripts/config/init.py --design-dir " + base_path + " --design-name " + design_name + " --add-to-designs" + "'"
      printc.cyan("Run config generation command", script_name=script_name)
      printc.bold(" > " + command)
      print("", end="", flush=True)
      process = subprocess.Popen(command, shell=True)
      while process.poll() is None:
        # print('.', end='', flush=True)
        time.sleep(0.5)
      print()
    except Exception as e:
      printc.error("Config file generation failed", script_name=script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)
      
  if not os.path.isfile(config_file):
    printc.error("Config file generation failed", script_name=script_name)
    sys.exit(-1)

  with open(config_file, 'r') as f:
    data = json.load(f)

  data['DESIGN_NAME'] = design_name
  data['CLOCK_PORT'] = clock_signal
  data['VERILOG_FILES'] = "dir::rtl/*.v"

  try: 
    with open(config_file, 'w') as f:
      json.dump(data, f, indent=4)
  except Exception as e:
    printc.error("Could not write config file \"" + config_file + "\"", script_name=script_name)
    printc.cyan("error details: ", end="", script_name=script_name)
    print(str(e))
    sys.exit(-1)

if __name__ == "__main__":
  args = parse_arguments()

  yaml_config_file = os.path.realpath(os.path.join(args.basepath, yaml_config_filename))
  arch = Architecture.read_yaml(yaml_config_file)

  if arch is None:
    printc.error("Could not read architecture settings from  \"" + yaml_config_file + "\"", script_name=script_name)
  else:
    if args.clock is None:
      args.clock = arch.clock_signal

    if args.docker is None:
      args.docker = arch.lib_name 

    if args.name is None:
      args.name = arch.top_level_module 

  if args.clock is None:
    printc.error("Missing --clock option", script_name=script_name)
    sys.exit(-1)

  if args.docker is None:
    printc.error("Missing --docker option", script_name=script_name)
    sys.exit(-1)

  if args.name is None:
    printc.error("Missing --name option", script_name=script_name)
    sys.exit(-1)

  main(
    base_path = args.basepath,
    clock_signal = args.clock,
    docker_id = args.docker,
    design_name = args.name
  )


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import math
import yaml

from os.path import isfile
from os.path import isdir

from odatix.lib.utils import *
import odatix.lib.printc as printc

script_name = os.path.basename(__file__)

odatix_path_pattern = re.compile(r"\$odatix")

class Architecture:
  def __init__(self, arch_name, arch_display_name, lib_name, target, tmp_script_path, tmp_report_path, tmp_dir, design_path, rtl_path, log_path, arch_path,
               clock_signal, reset_signal, top_level_module, top_level_filename, use_parameters, start_delimiter, stop_delimiter,
               file_copy_enable, file_copy_source, file_copy_dest, script_copy_enable, script_copy_source, 
               fmax_lower_bound, fmax_upper_bound, param_target_filename, generate_rtl, generate_command, constraint_filename):
    self.arch_name = arch_name
    self.arch_display_name = arch_display_name
    self.lib_name = lib_name
    self.target = target
    self.tmp_script_path = tmp_script_path
    self.tmp_report_path = tmp_report_path
    self.tmp_dir = tmp_dir
    self.design_path = design_path
    self.rtl_path = rtl_path
    self.log_path = log_path
    self.arch_path = arch_path
    self.clock_signal = clock_signal
    self.reset_signal = reset_signal
    self.top_level_module = top_level_module
    self.top_level_filename = top_level_filename
    self.file_copy_enable = file_copy_enable
    self.file_copy_source = file_copy_source
    self.file_copy_dest = file_copy_dest
    self.script_copy_enable = script_copy_enable
    self.script_copy_source = script_copy_source
    self.fmax_lower_bound = fmax_lower_bound
    self.fmax_upper_bound = fmax_upper_bound
    self.param_target_filename = param_target_filename
    self.use_parameters = use_parameters
    self.start_delimiter = start_delimiter
    self.stop_delimiter = stop_delimiter
    self.generate_rtl = generate_rtl
    self.generate_command = generate_command
    self.constraint_filename = constraint_filename

  def write_yaml(arch, config_file): 
    yaml_data = {
      'arch_name': arch.arch_name,
      'arch_display_name': arch.arch_display_name,
      'lib_name': arch.lib_name,
      'target': arch.target,
      'script_path': arch.tmp_script_path,
      'report_path': arch.tmp_report_path,
      'tmp_path': arch.tmp_dir,
      'design_path': arch.design_path,
      'rtl_path': arch.rtl_path,
      'log_path': arch.log_path,
      'arch_path': arch.arch_path,
      'clock_signal': arch.clock_signal,
      'reset_signal': arch.reset_signal,
      'top_level_module': arch.top_level_module,
      'top_level_file': arch.top_level_filename,
      'use_parameters': arch.use_parameters,
      'start_delimiter': arch.start_delimiter,
      'stop_delimiter': arch.stop_delimiter,
      'file_copy_enable': arch.file_copy_enable,
      'file_copy_source': arch.file_copy_source,
      'file_copy_dest': arch.file_copy_dest,
      'script_copy_enable': arch.script_copy_enable,
      'script_copy_source': arch.script_copy_source,
      'fmax_lower_bound': arch.fmax_lower_bound,
      'fmax_upper_bound': arch.fmax_upper_bound,
      'param_target_filename': arch.param_target_filename,
      'generate_rtl': arch.generate_rtl,
      'generate_command': arch.generate_command,
      'constraint_filename': arch.constraint_filename
    }
      
    with open(config_file, 'w') as f:
      yaml.dump(yaml_data, f, default_flow_style=False, sort_keys=False)

  def read_yaml(config_file):
    if not os.path.isfile(config_file):
      printc.error("Settings file \"" + config_file + "\" does not exist", script_name)
      return None

    with open(config_file, 'r') as f:
      yaml_data = yaml.safe_load(f)
    
    try:
      arch = Architecture(
        arch_name             = read_from_list("arch_name", yaml_data, config_file, script_name=script_name),
        arch_display_name     = read_from_list("arch_display_name", yaml_data, config_file, script_name=script_name),
        lib_name              = read_from_list("lib_name", yaml_data, config_file, script_name=script_name),
        target                = read_from_list("target", yaml_data, config_file, script_name=script_name),
        tmp_script_path       = read_from_list("script_path", yaml_data, config_file, script_name=script_name),
        tmp_report_path       = read_from_list("report_path", yaml_data, config_file, script_name=script_name),
        tmp_dir               = read_from_list("tmp_path", yaml_data, config_file, script_name=script_name),
        design_path           = read_from_list("design_path", yaml_data, config_file, script_name=script_name),
        rtl_path              = read_from_list("rtl_path", yaml_data, config_file, script_name=script_name),
        log_path              = read_from_list("log_path", yaml_data, config_file, script_name=script_name),
        arch_path             = read_from_list("arch_path", yaml_data, config_file, script_name=script_name),
        clock_signal          = read_from_list("clock_signal", yaml_data, config_file, script_name=script_name),
        reset_signal          = read_from_list("reset_signal", yaml_data, config_file, script_name=script_name),
        top_level_module      = read_from_list("top_level_module", yaml_data, config_file, script_name=script_name),
        top_level_filename    = read_from_list("top_level_file", yaml_data, config_file, script_name=script_name),
        use_parameters        = read_from_list("use_parameters", yaml_data, config_file, script_name=script_name),
        start_delimiter       = read_from_list("start_delimiter", yaml_data, config_file, script_name=script_name),
        stop_delimiter        = read_from_list("stop_delimiter", yaml_data, config_file, script_name=script_name),
        file_copy_enable      = read_from_list("file_copy_enable", yaml_data, config_file, script_name=script_name),
        file_copy_source      = read_from_list("file_copy_source", yaml_data, config_file, script_name=script_name),
        file_copy_dest        = read_from_list("file_copy_dest", yaml_data, config_file, script_name=script_name),
        script_copy_enable    = read_from_list("script_copy_enable", yaml_data, config_file, script_name=script_name),
        script_copy_source    = read_from_list("script_copy_source", yaml_data, config_file, script_name=script_name),
        fmax_lower_bound      = read_from_list("fmax_lower_bound", yaml_data, config_file, script_name=script_name),
        fmax_upper_bound      = read_from_list("fmax_upper_bound", yaml_data, config_file, script_name=script_name),
        param_target_filename = read_from_list("param_target_filename", yaml_data, config_file, script_name=script_name),
        generate_rtl          = read_from_list("generate_rtl", yaml_data, config_file, script_name=script_name),
        generate_command      = read_from_list("generate_command", yaml_data, config_file, script_name=script_name),
        constraint_filename   = read_from_list("constraint_filename", yaml_data, config_file, script_name=script_name)
      )
    except (KeyNotInListError, BadValueInListError):
      return None
    return arch

class ArchitectureHandler:

  def __init__(self, work_path, arch_path, script_path, work_script_path, work_report_path, log_path, process_group, eda_target_filename, fmax_status_filename, frequency_search_filename, param_settings_filename, valid_status, valid_frequency_search, default_fmax_lower_bound, default_fmax_upper_bound, overwrite):
    self.work_path = work_path
    self.arch_path = arch_path
    self.script_path = script_path
    self.work_script_path = work_script_path
    self.work_report_path = work_report_path
    self.log_path = log_path

    self.process_group = process_group
    
    self.eda_target_filename = eda_target_filename
    self.fmax_status_filename = fmax_status_filename
    self.frequency_search_filename = frequency_search_filename
    self.param_settings_filename = param_settings_filename
    
    self.valid_status = valid_status
    self.valid_frequency_search = valid_frequency_search

    self.default_fmax_lower_bound = default_fmax_lower_bound
    self.default_fmax_upper_bound = default_fmax_upper_bound

    self.overwrite = overwrite
    self.reset_lists()

    self.odatix_path = os.path.realpath(os.path.join(self.script_path, ".."))

  def reset_lists(self):
    self.banned_arch_param = []
    self.valid_archs = []
    self.cached_archs = []
    self.overwrite_archs = []
    self.error_archs = []
    self.incomplete_archs = []
    self.new_archs = []

  def get_architectures(self, architectures, targets, constraint_filename=""):

    self.reset_lists()
    self.architecture_instances = []

    only_one_target = len(targets) == 1
      
    with open(self.eda_target_filename, 'r') as f:
      try:
        settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
      except Exception as e:
        printc.error("Settings file \"" + self.eda_target_filename + "\" is not a valid YAML file", script_name)
        printc.cyan("error details: ", end="", script_name=script_name)
        print(str(e))
        sys.exit(-1)
      
      try:
        script_copy_enable = read_from_list('script_copy_enable', settings_data, self.eda_target_filename, type=bool, optional=True, script_name=script_name)
        if script_copy_enable:
          script_copy_source = read_from_list('script_copy_source', settings_data, self.eda_target_filename, optional=True, script_name=script_name)        
          script_copy_source = os.path.realpath(re.sub(odatix_path_pattern, self.odatix_path, script_copy_source))
          if not os.path.isfile(script_copy_source):
            printc.note("The script source file \"" + script_copy_source + "\" specified in \"" + self.eda_target_filename + "\" does not exist. Script copy disabled.", script_name)
            raise BadValueInListError
        else:
          raise BadValueInListError
      except (KeyNotInListError, BadValueInListError):
        script_copy_enable = False
        script_copy_source = "/dev/null"

      try:
        target_settings = read_from_list("target_settings", settings_data, self.eda_target_filename, optional=True, print_error=False, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        target_settings = {}

      for target in targets:
        # Overwrite existing script copy settings if there are target specific settings
        if target_settings != {}:
          try:
            this_target_settings = read_from_list(target, target_settings, self.eda_target_filename, optional=True, parent="target_settings", script_name=script_name)
          except (KeyNotInListError, BadValueInListError):
            this_target_settings = {}
            pass
          if this_target_settings != {}:
            try:
              script_copy_enable = read_from_list('script_copy_enable', this_target_settings, self.eda_target_filename, type=bool, optional=True, parent="target_settings/" + target, script_name=script_name)
              if script_copy_enable:
                script_copy_source = read_from_list('script_copy_source', this_target_settings, self.eda_target_filename, optional=True, parent="target_settings/" + target, script_name=script_name)        
                script_copy_source = os.path.realpath(re.sub(odatix_path_pattern, self.odatix_path, script_copy_source))

                if not os.path.isfile(script_copy_source):
                  printc.note("The script source file \"" + script_copy_source + "\" specified in \"" + self.eda_target_filename + "\" does not exist. Script copy disabled.", script_name)
                  raise BadValueInListError
            except (KeyNotInListError, BadValueInListError):
              script_copy_enable = False
              script_copy_source = "/dev/null"

        for arch in architectures:
          architecture_instance = self.get_architecture(
            arch = arch,
            target = target, 
            only_one_target = only_one_target, 
            script_copy_enable = script_copy_enable, 
            script_copy_source = script_copy_source,
            synthesis = True,
            constraint_filename = constraint_filename
          )
          if architecture_instance is not None:
            self.architecture_instances.append(architecture_instance)

    return self.architecture_instances
  
  def get_architecture(self, arch, target="", only_one_target=True, script_copy_enable=False, script_copy_source="/dev/null", synthesis=False, constraint_filename=""):

    if arch.endswith(".txt"):
      arch = arch[:-4] 
      printc.note("'.txt' after the configuration name is not needed. Just use \"" + arch + "\"", script_name)

    if arch.endswith("/"):
      arch = arch[:-1] 

    if only_one_target:
      arch_display_name = arch
    else:
      arch_display_name = arch + " (" + target + ")"

    # get param dir (arch name before '/')
    arch_param_dir = re.sub('/.*', '', arch)

    # get param dir (arch name after '/')
    arch_suffix = re.sub('.*/', '', arch)

    # check if there is a configuration specified
    if arch_suffix == arch_param_dir:
      printc.note("No architecture configuration selected for \"" + arch +  "\". Using default parameters", script_name)
      arch = arch + "/" + arch
      no_configuration = True
    else:
      no_configuration = False

    tmp_dir = self.work_path + '/' + target + '/' + arch
    fmax_status_file = tmp_dir + '/' + self.log_path + '/' + self.fmax_status_filename
    frequency_search_file = tmp_dir + '/' + self.log_path + '/' + self.frequency_search_filename

    # check if arch_param has been banned
    if arch_param_dir in self.banned_arch_param:
      self.error_archs.append(arch_display_name)
      return None

    # check if parameter dir exists
    arch_param = self.arch_path + '/' + arch_param_dir
    if not isdir(arch_param):
      printc.error("There is no directory \"" + arch_param_dir + "\" in directory \"" + self.arch_path + "\"", script_name)
      self.banned_arch_param.append(arch_param_dir)
      self.error_archs.append(arch_display_name)
      return None
    
    # check if settings file exists
    if not isfile(arch_param + '/' + self.param_settings_filename):
      printc.error("There is no setting file \"" + self.param_settings_filename + "\" in directory \"" + arch_param + "\"", script_name)
      self.banned_arch_param.append(arch_param_dir)
      self.error_archs.append(arch_display_name)
      return None

    # get settings variables
    settings_filename = self.arch_path + '/' + arch_param_dir + '/' + self.param_settings_filename
    with open(settings_filename, 'r') as f:
      try:
        settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
      except Exception as e:
        printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
        printc.cyan("error details: ", end="", script_name=script_name)
        print(str(e))
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None # if an identifier is missing
      try:
        rtl_path           = read_from_list('rtl_path', settings_data, settings_filename, script_name=script_name)
        top_level_filename = read_from_list('top_level_file', settings_data, settings_filename, script_name=script_name)
        top_level_module   = read_from_list('top_level_module', settings_data, settings_filename, script_name=script_name)
        clock_signal       = read_from_list('clock_signal', settings_data, settings_filename, script_name=script_name)
        reset_signal       = read_from_list('reset_signal', settings_data, settings_filename, script_name=script_name)
        file_copy_enable   = read_from_list('file_copy_enable', settings_data, settings_filename, type=bool, script_name=script_name)
        file_copy_source   = read_from_list('file_copy_source', settings_data, settings_filename, script_name=script_name)
        file_copy_dest     = read_from_list('file_copy_dest', settings_data, settings_filename, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None # if an identifier is missing

      use_parameters, start_delimiter, stop_delimiter = self.get_use_parameters(arch, arch_display_name, settings_data, settings_filename, no_configuration, arch_param_dir=arch_param_dir)
      if use_parameters is None or start_delimiter is None or stop_delimiter is None:
        return None

      generate_command = ""
      try:
        generate_rtl = read_from_list('generate_rtl', settings_data, settings_filename, type=bool, optional=True, print_error=False, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        generate_rtl = False

      if generate_rtl:
        try:
          generate_command = read_from_list('generate_command', settings_data, settings_filename, print_error=False, script_name=script_name)
          generate_rtl = True
        except (KeyNotInListError, BadValueInListError):
          printc.error("Cannot find key \"generate_command\" in \"" + settings_filename + "\" while generate_rtl=true", script_name)
          self.banned_arch_param.append(arch_param_dir)
          self.error_archs.append(arch_display_name)
          generate_rtl = False
          return None

      try:
        design_path = read_from_list('design_path', settings_data, settings_filename, optional=True, print_error=False, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        design_path = -1
        if generate_rtl:
          printc.error("Cannot find key \"design_path\" in \"" + settings_filename + "\" while generate_rtl=true", script_name)
          self.banned_arch_param.append(arch_param_dir)
          return None
      
      try:
        param_target_filename = read_from_list('param_target_file', settings_data, settings_filename, optional=True, print_error=False, script_name=script_name)
        if design_path == -1:
          printc.error("Cannot find key \"design_path\" in \"" + settings_filename + "\" while param_target_file is defined", script_name)
          self.banned_arch_param.append(arch_param_dir)
          return None
        # check if param target file path exists
        param_target_file = design_path + '/' + param_target_filename
        if not isfile(param_target_file): 
          printc.error("The parameter target file \"" + param_target_filename + "\" specified in \"" + settings_filename + "\" does not exist", script_name)
          self.banned_arch_param.append(arch_param_dir)
          return None
      except (KeyNotInListError, BadValueInListError):
        param_target_filename = 'rtl/' + top_level_filename

    if not generate_rtl:
      # check if rtl path exists
      if not isdir(rtl_path):
        printc.error("The rtl path \"" + rtl_path + "\" specified in \"" + settings_filename + "\" does not exist", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None

      # check if top level file path exists
      top_level = rtl_path + '/' + top_level_filename
      if not isfile(top_level):
        printc.error("The top level file \"" + top_level_filename + "\" specified in \"" + settings_filename + "\" does not exist", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None

      # check if the top level module name exists in the top level file, at least
      f = open(top_level, "r")
      if top_level_module not in f.read():
        printc.error("There is no occurence of top level module name \"" + top_level_module + "\" in top level file \"" + top_level + "\"", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        f.close()
        return None
      f.close()
      
      # check if the top clock name exists in the top level file, at least
      f = open(top_level, "r")
      if clock_signal not in f.read():
        printc.error("There is no occurence of clock signal name \"" + clock_signal + "\" in top level file \"" + top_level + "\"", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        f.close()
        return None
      f.close()
      
      # check if the top reset name exists in the top level file, at least
      f = open(top_level, "r")
      if clock_signal not in f.read():
        printc.error("There is no occurence of reset signal name \"" + reset_signal + "\" in top level file \"" + top_level + "\"", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        f.close()
        return None
      f.close()

    # check if param file exists
    if not no_configuration:
      if not isfile(self.arch_path + '/' + arch + '.txt'):
        printc.error("The parameter file \"" + arch + ".txt\" does not exist in directory \"" + self.arch_path + "/" + arch_param_dir + "\"", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None

    # optional settings
    formatted_bound = ""
    fmax_lower_bound_ok = False
    fmax_upper_bound_ok = False
    fmax_lower_bound = 0
    fmax_upper_bound = 0

    target_options = read_from_list(target, settings_data, settings_filename, optional=True, raise_if_missing=False, print_error=False, script_name=script_name)
    
    if synthesis:
      if target_options == False:
        printc.note("Cannot find optional target-specific options for target \"" + target + "\" in \"" + settings_filename + "\". Using default frequency bounds instead: " + "[{},{}] MHz.".format(self.default_fmax_lower_bound, self.default_fmax_upper_bound), script_name)
        fmax_lower_bound = self.default_fmax_lower_bound
        fmax_upper_bound = self.default_fmax_upper_bound
      else:
        architectures_bounds = read_from_list('architectures', target_options, self.eda_target_filename, optional=True, raise_if_missing=False, print_error=False, script_name=script_name)
        if architectures_bounds:
          this_architecture_bounds = read_from_list(arch_suffix, architectures_bounds, self.eda_target_filename, optional=True, raise_if_missing=False, print_error=False, script_name=script_name)
          if this_architecture_bounds:
            fmax_lower_bound = read_from_list('fmax_lower_bound', this_architecture_bounds, self.eda_target_filename, optional=True, raise_if_missing=False, print_error=False, script_name=script_name)
            if fmax_lower_bound:
              fmax_lower_bound_ok = True
            
            fmax_upper_bound = read_from_list('fmax_upper_bound', this_architecture_bounds, self.eda_target_filename, optional=True, raise_if_missing=False, print_error=False, script_name=script_name)
            if fmax_upper_bound:
              fmax_upper_bound_ok = True

            # check if bounds are valid
            if (fmax_upper_bound <= fmax_lower_bound) : 
              printc.error("The upper bound (" + fmax_upper_bound + ") must be strictly superior to the lower bound (" + fmax_lower_bound + ")", script_name)
              self.error_archs.append(arch_display_name)
              return None

        if fmax_lower_bound_ok == False:
          fmax_lower_bound = read_from_list('fmax_lower_bound', target_options, self.eda_target_filename, optional=True, raise_if_missing=False, script_name=script_name)
          if fmax_lower_bound == False:
            printc.note("Cannot find optional key \"fmax_lower_bound\" for target \"" + target + "\" in \"" + settings_filename + "\". Using default frequency lower bound instead: " + "{} MHz.".format(self.default_fmax_lower_bound), script_name)
            fmax_lower_bound = self.default_fmax_lower_bound
          #else:
            #printc.note("Cannot find optional key \"fmax_lower_bound\" for architecture \"" + arch + "\" with target \"" + target + "\" in \"" + settings_filename + "\". Using target frequency lower bound instead: " + "{} MHz.".format(self.default_fmax_lower_bound), script_name)

        if fmax_upper_bound_ok == False:
          fmax_upper_bound = read_from_list('fmax_upper_bound', target_options, self.eda_target_filename, optional=True, raise_if_missing=False, script_name=script_name)
          if fmax_upper_bound == False:
            printc.note("Cannot find optional key \"fmax_upper_bound\" for target \"" + target + "\" in \"" + settings_filename + "\". Using default frequency upper bound instead: " + "{} MHz.".format(self.default_fmax_upper_bound), script_name)
            fmax_upper_bound = self.default_fmax_upper_bound
          #else:
            #printc.note("Cannot find optional key \"fmax_upper_bound\" for architecture \"" + arch + "\" with target \"" + target + "\" in \"" + settings_filename + "\". Using target frequency upper bound instead: " + "{} MHz.".format(self.default_fmax_upper_bound), script_name)
            
      # check if bounds are valid
      if (fmax_upper_bound <= fmax_lower_bound) : 
        printc.error("The upper bound (" + fmax_upper_bound + ") must be strictly superior to the lower bound (" + fmax_lower_bound + ")", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None

      fmax_lower_bound = str(fmax_lower_bound)
      fmax_upper_bound = str(fmax_upper_bound)

      formatted_bound = " {}({} - {} MHz){}".format(printc.colors.GREY, fmax_lower_bound, fmax_upper_bound, printc.colors.ENDC)

      # check if the architecture is in cache and has a status file
      if isdir(tmp_dir) and isfile(fmax_status_file) and isfile(frequency_search_file):
        # check if the previous synth_fmax has completed
        sf = open(fmax_status_file, "r")
        if self.valid_status in sf.read():
          ff = open(frequency_search_file, "r")
          if self.valid_frequency_search in ff.read():
            if self.overwrite:
              printc.warning("Found cached results for \"" + arch + "\" with target \"" + target + "\".", script_name)
              self.overwrite_archs.append(arch_display_name + formatted_bound)
            else:
              printc.note("Found cached results for \"" + arch + "\" with target \"" + target + "\". Skipping.", script_name)
              self.cached_archs.append(arch_display_name)
              return None
          else:
            printc.warning("The previous synthesis for \"" + arch + "\" did not result in a valid maximum operating frequency.", script_name)
            self.incomplete_archs.append(arch_display_name + formatted_bound)
          ff.close()
        else: 
          printc.warning("The previous synthesis for \"" + arch + "\" has not finished or the directory has been corrupted.", script_name)
          self.incomplete_archs.append(arch_display_name + formatted_bound)
        sf.close()
      else:
        self.new_archs.append(arch_display_name + formatted_bound)

    # target specific file copy
    if target_options:
      try:
        _file_copy_enable = read_from_list('file_copy_enable', target_options, self.eda_target_filename, optional=True, print_error=False, type=bool, script_name=script_name)
        try:
          _file_copy_source = read_from_list('file_copy_source', target_options, self.eda_target_filename, optional=True, script_name=script_name)
          _file_copy_dest = read_from_list('file_copy_dest', target_options, self.eda_target_filename, optional=True, script_name=script_name)
          file_copy_enable = _file_copy_enable
          file_copy_source = _file_copy_source
          file_copy_dest = _file_copy_dest
        except (KeyNotInListError, BadValueInListError):
          pass
      except KeyNotInListError:
        pass
      except BadValueInListError:
        printc.note("Value \"" + str(_file_copy_enable) + "\" for key \"" + 'file_copy_enable' + "\"" + ", inside list \"" + "target_settings/" + target + "\"," + " in \"" + self.eda_target_filename + "\" is of type \"" + _file_copy_enable.__class__.__name__ + "\" while it should be of type \"bool\". Using default values instead.", script_name)

    # check file copy
    file_copy_source = os.path.realpath(re.sub(odatix_path_pattern, self.odatix_path, file_copy_source))
    if file_copy_enable:
      if not isfile(file_copy_source):
        printc.error("The source file to copy \"" + file_copy_source + "\" does not exist", script_name)
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
        return None

    # passed all check: added to the list
    self.valid_archs.append(arch_display_name)

    lib_name = "LIB_" + target + "_" + arch.replace("/", "_")

    tmp_script_path = os.path.join(tmp_dir, self.work_script_path)
    tmp_report_path = os.path.join(tmp_dir, self.work_report_path)

    arch_instance = Architecture(
      arch_name = arch,
      arch_display_name = arch_display_name,
      lib_name = lib_name,
      target = target,
      tmp_script_path=tmp_script_path,
      tmp_report_path=tmp_report_path,
      tmp_dir=tmp_dir,
      design_path=design_path,
      rtl_path=rtl_path,
      log_path=self.log_path,
      arch_path=self.arch_path,
      clock_signal=clock_signal,
      reset_signal=reset_signal,
      top_level_module=top_level_module,
      top_level_filename=top_level_filename,
      file_copy_enable=file_copy_enable,
      file_copy_source=file_copy_source,
      file_copy_dest=file_copy_dest,
      script_copy_enable = script_copy_enable,
      script_copy_source = script_copy_source,
      fmax_lower_bound=fmax_lower_bound,
      fmax_upper_bound=fmax_upper_bound,
      param_target_filename=param_target_filename,
      generate_rtl=generate_rtl,
      use_parameters=use_parameters,
      start_delimiter=start_delimiter,
      stop_delimiter=stop_delimiter,
      generate_command=generate_command,
      constraint_filename=constraint_filename
    )

    return arch_instance

  def get_use_parameters(self, arch, arch_display_name, settings_data, settings_filename, no_configuration=False, add_to_error_list=True, arch_param_dir=""):
    # get use_parameters
    try:
      use_parameters = read_from_list('use_parameters', settings_data, settings_filename, type=bool, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      if add_to_error_list:
        self.banned_arch_param.append(arch_param_dir)
        self.error_archs.append(arch_display_name)
      return None, None, None

    param_filename = arch + ".txt"
    if no_configuration:
      use_parameters = False
    else:
      if use_parameters:
        # check if parameter file exists
        if not isfile(self.arch_path + '/' + param_filename):
          printc.error("There is no parameter file \"" + self.arch_path + '/' + param_filename + "\", while use_parameters=true", script_name)
          if add_to_error_list:
            self.error_archs.append(arch_display_name)
          return True, None, None
    
    if use_parameters:
      # get start delimiter
      try:
        start_delimiter = read_from_list('start_delimiter', settings_data, settings_filename, print_error=False, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        printc.error("Cannot find key \"start_delimiter\" in \"" + settings_filename + "\", while \"use_parameters\" is true", script_name)
        if add_to_error_list:
          self.banned_arch_param.append(arch_param_dir)
          self.error_archs.append(arch_display_name)
        return None, None, None

      # get stop delimiter
      try:
        stop_delimiter = read_from_list('stop_delimiter', settings_data, settings_filename, print_error=False, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        printc.error("Cannot find key \"stop_delimiter\" in \"" + settings_filename + "\", while \"use_parameters\" is true", script_name)
        if add_to_error_list:
          self.banned_arch_param.append(arch_param_dir)
          self.error_archs.append(arch_display_name)
        return None, None, None
    else:
      start_delimiter = ""
      stop_delimiter = ""

    return use_parameters, start_delimiter, stop_delimiter

  def print_summary(self):
    ArchitectureHandler.print_arch_list(self.new_archs, "New architectures", printc.colors.ENDC)
    ArchitectureHandler.print_arch_list(self.incomplete_archs, "Incomplete results (will be overwritten)", printc.colors.YELLOW)
    ArchitectureHandler.print_arch_list(self.cached_archs, "Existing results (skipped)", printc.colors.CYAN)
    ArchitectureHandler.print_arch_list(self.overwrite_archs, "Existing results (will be overwritten)", printc.colors.YELLOW)
    ArchitectureHandler.print_arch_list(self.error_archs, "Invalid settings, (skipped, see errors above)", printc.colors.RED)

  def get_chuncks(self, nb_jobs):
    if len(self.architecture_instances) > nb_jobs:
      nb_chunks = math.ceil(len(self.architecture_instances) / nb_jobs)
      print()
      printc.note("Current maximum number of jobs is " + str(nb_jobs) + ". Architectures will be split in " + str(nb_chunks) + " chunks")
      self.architecture_instances_chunks = list(chunk_list(self.architecture_instances, nb_jobs))
    else:
      nb_chunks = 1
      self.architecture_instances_chunks = []
    return self.architecture_instances_chunks, nb_chunks


  def get_valid_arch_count(self):
    return len(self.valid_archs)

  @staticmethod
  def print_arch_list(arch_list, description, color):
    if not len(arch_list) > 0:
      return

    print()
    printc.bold(description + ":")
    for arch in arch_list:
      printc.color(color)
      print("  - " + arch)
    printc.endc()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import plotly.express as px
import re

plot_colors = px.colors.qualitative.Plotly


def create_legend_items(explorer, page_name=""):
  legend_items = [
    create_legend_item(
      architecture=architecture, line_style="2px dashed", color=plot_colors[i % len(plot_colors)], page_name=page_name
    )
    for i, architecture in enumerate(explorer.all_architectures)
  ]
  return legend_items


def create_legend_item(architecture, line_style, color, page_name=""):
  return html.Div(
    [
      dcc.Checklist(
        id=f"checklist-{architecture}-{page_name}",
        options=[{"label": "", "value": architecture}],
        value=[architecture],
        inline=True,
        style={"display": "inline-block", "margin-right": "10px", "text-wrap": "wrap"},
      ),
      html.Div(
        style={
          "display": "inline-block",
          "width": "30px",
          "height": "2px",
          "border-top": f"{line_style} {color}",
          "position": "relative",
        },
        children=html.Div(
          style={
            "position": "absolute",
            "top": "-6px",
            "left": "50%",
            "transform": "translateX(-50%)",
            "width": "10px",
            "height": "10px",
            "background-color": color,
            "border-radius": "50%",
          }
        ),
      ),
      html.Div(f"{architecture}", style={"display": "inline-block", "margin-left": "5px"}),
    ],
    id=f"legend-item-{architecture}-{page_name}",
    style={"display": "block", "margin-bottom": "5px"},
  )


def setup_callbacks(explorer, page_name):
  @explorer.app.callback(
    [Output(f"checklist-{architecture}-{page_name}", "value") for architecture in explorer.all_architectures],
    [Input("show-all", "n_clicks"), Input("hide-all", "n_clicks")],
    [State(f"checklist-{architecture}-{page_name}", "value") for architecture in explorer.all_architectures],
  )
  def update_checklist_states(show_all_clicks, hide_all_clicks, *current_values):
    ctx = dash.callback_context
    if not ctx.triggered:
      return dash.no_update

    button_id = ctx.triggered[0]["prop_id"].split(".")[0]
    if button_id == "show-all":
      return [[architecture] for architecture in explorer.all_architectures]
    elif button_id == "hide-all":
      return [[] for _ in explorer.all_architectures]

    return current_values

  @explorer.app.callback(
    [Output(f"legend-item-{architecture}-{page_name}", "style") for architecture in explorer.all_architectures],
    [Input(f"target-dropdown-{page_name}", "value"), Input("yaml-dropdown", "value")],
  )
  def update_legend_visibility(selected_target, selected_yaml):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return [{"display": "none"} for _ in explorer.all_architectures]

    architectures_for_target = explorer.dfs[selected_yaml][explorer.dfs[selected_yaml]["Target"] == selected_target][
      "Architecture"
    ].unique()
    return [
      {"display": "block" if architecture in architectures_for_target else "none"}
      for architecture in explorer.all_architectures
    ]


def get_color(i):
  return plot_colors[i % len(plot_colors)]

def unit_to_html(unit):
  # Regex pattern to match ^(-?\d+) for positive/negative exponents
  pattern = r'\^(-?\d+)'
  html_unit = re.sub(pattern, r'<sup>\1</sup>', unit)

  # Regex pattern to match _(-?\d+) for positive/negative subscripts
  pattern_sub = r'\_(-?\d+)'
  unit_with_sub = re.sub(pattern_sub, r'<sub>\1</sub>', html_unit)

  return unit_with_sub


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re

BAD_VALUE = ' /   '

def get_re_group_from_file(file, pattern, group_id, bad_value=BAD_VALUE):
  if os.path.exists(file):
    for i, line in enumerate(open(file)):
      for match in re.finditer(pattern, line):
        parts = pattern.search(match.group())
        if group_id <= len(parts.groups()):
          return parts.group(group_id)
  return bad_value



import os
from setuptools import setup, find_packages

def read_version():
    with open("odatix/version.txt", "r") as file:
        return file.read().strip()

def read_requirements():
    with open("requirements.txt") as f:
        return [line.strip() for line in f if line.strip()]

package_list = find_packages()
package_list.append('odatix_examples')
package_list.append('odatix_eda_tools')
package_list.append('odatix_init')

setup(
    name="odatix",
    version=read_version(),
    author="Jonathan Saussereau",
    author_email="jonathan.saussereau@ims-bordeaux.fr",
    description="A FPGA/ASIC toolbox for design space exploration",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jsaussereau/Asterism",
    packages=package_list,
    package_data={
        'odatix': [
            'version.txt',
            'explorer/assets/**/*',
            'init/**/*',
            'eda_tools/**/*',
        ],
        'odatix_examples': [
            '**/*'
        ],
        'odatix_eda_tools': [
            '**/*'
        ],
        'odatix_init': [
            '**/*'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=read_requirements(),
    entry_points={
        "console_scripts": [
            "odatix=odatix.odatix_main:main",
            "odatix-explorer=odatix.odatix_explorer:main",
        ],
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import argparse

import odatix.lib.printc as printc

script_name = os.path.basename(__file__)

def read_file(file_path):
    try:
        with open(file_path, 'r') as file:
            content = file.read()
        return content
    except:
        printc.error("could not open input file \"" + file_path + "\"", script_name)
        sys.exit(1)

def write_file(file_path, content):
    try:
        with open(file_path, 'w') as file:
            file.write(content)
    except Exception as e:
        printc.error("could not write output file \"" + file_path + "\": " + str(e), script_name)
        sys.exit(1)

def replace_content(base_text, replacement_text, start_delim, stop_delim, replace_all_occurrences):
    pattern = re.escape(start_delim) + '.*?' + re.escape(stop_delim)

    match_found = re.search(pattern, base_text, flags=re.DOTALL) is not None
    
    if replace_all_occurrences:
        new_text = re.sub(pattern, start_delim + replacement_text + stop_delim, base_text, flags=re.DOTALL)
    else:
        new_text = re.sub(pattern, start_delim + replacement_text + stop_delim, base_text, count=1, flags=re.DOTALL)
    return new_text, match_found

def replace_params(base_text_file, replacement_text_file, output_file, start_delimiter, stop_delimiter, replace_all_occurrences=False, silent=False):
    # Read the contents of text files
    base_text = read_file(base_text_file)
    replacement_text = read_file(replacement_text_file)

    # Replace content between delimiters
    new_text, match_found = replace_content(base_text, replacement_text, start_delimiter, stop_delimiter, replace_all_occurrences)

    # Write the new content to the output file
    write_file(output_file, new_text)

    if not match_found:
        printc.warning("could not find pattern \"", script_name=script_name, end="")
        printc.red(start_delimiter, end="")
        printc.grey("[…]", end="")
        printc.red(stop_delimiter, end="")
        printc.yellow("\" in \"" + base_text_file + "\"")

    if not silent and match_found:
        if new_text != base_text:
            printc.say("content replaced successfully, output saved to \"" + output_file + "\"", script_name)
        else:
            printc.say("nothing to be done, input copied to \"" + output_file + "\"", script_name)

def parse_arguments():
    parser = argparse.ArgumentParser(description="Replace content between delimiters in a text file.")
    parser.add_argument("-s", "--startdel", dest="start_delimiter", required=True, help="Start delimiter")
    parser.add_argument("-S", "--stopdel", dest="stop_delimiter", required=True, help="Stop delimiter")
    parser.add_argument("-i", "--input", dest="base_text_file", required=True, help="Input base text file")
    parser.add_argument("-o", "--output", dest="output_file", required=True, help="Output text file")
    parser.add_argument("-r", "--replace", dest="replacement_text_file", required=True, help="Replacement text file")
    parser.add_argument("-a", "--all", dest="replace_all_occurrences", action="store_true", help="Replace all occurrences")
    return parser.parse_args()


######################################
# Main
######################################

def main():
    args = parse_arguments()
    replace_params(args.base_text_file, args.replacement_text_file, args.output_file, args.start_delimiter, args.stop_delimiter, args.replace_all_occurrences)

if __name__ == "__main__":
    main()


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import argparse

import odatix.lib.printc as printc

current_dir = os.path.dirname(os.path.abspath(__file__))

# get eda_tools folder
if getattr(sys, 'frozen', False):
  base_path = os.path.dirname(sys.executable)
  
else:
  base_path = current_dir

######################################
# Settings
######################################

version_file = os.path.realpath(os.path.join(base_path, os.pardir, "version.txt"))

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument('-f', '--full', action='store_true', help='show full motd')

def parse_arguments():
  parser = argparse.ArgumentParser(description="Odatix's message of the day")
  add_arguments(parser)
  return parser.parse_args()

######################################
# Message of the day
######################################

def motd():
  try:
    print()
    print(" ██████╗  ██████╗   █████╗ ████████╗ ██╗ ██╗  ██╗")
    print("██╔═══██╗ ██╔══██╗ ██╔══██╗╚══██╔══╝ ██║ ╚██╗██╔╝")
    print("██║   ██║ ██║  ██║ ███████║   ██║    ██║  ╚███╔╝ ")
    print("██║   ██║ ██║  ██║ ██╔══██║   ██║    ██║  ██╔██╗ ")
    print("╚██████╔╝ ██████╔╝ ██║  ██║   ██║    ██║ ██╔╝ ██╗")
    print(" ╚═════╝  ╚═════╝  ╚═╝  ╚═╝   ╚═╝    ╚═╝ ╚═╝  ╚═╝")   
  except:
    print()
    print(r" ######\         ##\              ##\      ##\            ")
    print(r"##  __##\        ## |             ## |     \__|           ")
    print(r"## /  ## |  ####### |  ######\  ######\    ##\  ##\   ##\ ")
    print(r"## |  ## | ##  __## |  \____##\ \_##  _|   ## | \##\ ##  |")
    print(r"## |  ## | ## /  ## |  ####### |  ## |     ## |  \####  / ")
    print(r"## |  ## | ## |  ## | ##  __## |  ## |##\  ## |  ##  ##\  ")
    print(r" ######  | \####### | \####### |  \####  | ## | ##  /\##\ ")
    print(r" \______/   \_______|  \_______|   \____/  \__| \__/  \__|")
    print()

def full_header(description=True):
  motd()
  if description:
    print("Odatix - a FPGA/ASIC toolbox for design space exploration")
  print_copyright()
  print()
  print("version: " + str(read_version()))
  print()

def print_copyright():
  print("Copyright (C) 2022-2024 Jonathan Saussereau")

######################################
# Version
######################################

def read_version():
  try:
    with open(version_file, 'r') as file:
      return file.read().strip()
  except:
    printc.warning("Could not read version file \"" + version_file + "\"")
    return "Unknown"

def print_version():
  version = read_version()
  print("Odatix " + str(version))

######################################
# Main
######################################

if __name__ == "__main__":
  args = parse_arguments()
  
  if args.full:
    full_header()
  else:
    motd()
